import z3
import z3_utils
import hcp_utils
import hc_problem as hcp
import dfs
import py_utils

import spacer_args

from stats import time_stats

def extend_signature (old_pred, num, sort) :
  """create a new pred from old_pred by adding `num' parameters of
`sort_kind' in `z3ctx'
  """
  new_sorts = list ()
  # begin with extra sorts to add
  new_sorts.extend ([sort for i in range (num)])
  # original domain sorts of old_pred
  new_sorts.extend ([z3_utils.z3_translate (old_pred.domain (i), sort.ctx) \
                     for i in range (old_pred.arity ())])
  # original range sort of old_pred
  new_sorts.append (z3_utils.z3_translate (old_pred.range (), sort.ctx))
  return z3.Function (old_pred.name () + '_b', *new_sorts)

@time_stats
def create_bounded_hc_problem (orig_prob, sort) :
  """bound pts by adding extra parameters and add new constraints to the rules
to bound the number of iterations of loops

  it's a skeleton as the bounding is parameterized by constants
denoting upper bounds on the number of iterations; different
instantiations of these constants give different bounded problems
  """

  assert orig_prob.ctx == sort.ctx

  # map from old pts to new pts
  bd_pts = dict ()

  assert orig_prob.closed
  bd_prob = hcp.HornClauseProblem (sort.ctx) # bounded skeleton to be created

  # create new preds by walking over the wto of orig_prob;
  bd_prob.wto = None

  for wto_comps, pt in hcp_utils.wtoWalk (orig_prob.wto) :
    bd_pred = extend_signature (pt.head, len (wto_comps), sort)
    bd_pt = bd_prob.ensure_pt (bd_pred)
    bd_pt.set_origin_pt (pt)
    bd_pt.set_id (pt.get_id ())
    bd_pt.set_num_bound_params (len (wto_comps))
    bd_pts [pt] = bd_pt

    begin_component = (len (wto_comps) > 0) and (pt == wto_comps [-1] [0])
    # create bound value const for a head
    if begin_component:  bd_pt.create_bound_const (sort)


  # create bounded rules by a second walk over wto
  for wto_comps, pt in hcp_utils.wtoWalk (orig_prob.wto) :

    bd_pt = bd_pts [pt]
    # for every rule defining pt
    for src,rule in hcp_utils.srcRulesWalk (pt) :

      bd_src = bd_pts.get (src)
      if bd_src is None :
        assert src is None
        src_pred = None
      else: src_pred = bd_src.head

      # Get bd_rule and create if not there
      bd_rule = bd_pt.get_rule_from (bd_src)
      if bd_rule is None :
        bd_rule = hcp.BoundedRule (rule.get_origin (), bd_pt.head, src_pred)
        bd_prob.add_rule (bd_rule)

      # special case of entry node
      if len(wto_comps) > 0: assert bd_src is not None


      ### COMPUTE PASS-ON CONSTRAINTS
      
      # edge within a wto component
      passon_cons = list ()

      # constraints for bound variables to be remembered, i.e. passed on
      # note that this rule has at most len(wto_comps) such params
      for i in range (len (wto_comps)):
        comp = wto_comps [i]
        # component's head
        hd = comp [0]
        
        # last case, nothing to pass on
        if hd == pt: continue
        
        # if src has the counter for me to pass on, do it
        if bd_src.get_num_bound_params () > i:
          pre_const = bd_src.pre_const (i)
          post_const = bd_pt.post_const (i)
          passon_cons.append (pre_const == post_const)
        else:
          # otherwise, leave counter non-deterministic
          pass

      # store constraints in bd_rule
      bd_rule.add_bound_cons (passon_cons)
      del passon_cons

      ### COMPUTE RE-ENTRY CONSTRAINTS


      # true iff pt is the head of the top-most component
      #  of the stack wto_comps
      pt_is_a_head = (len (wto_comps) > 0) and (pt == wto_comps [-1] [0])

      # add reentry constraints
      if pt_is_a_head and hcp_utils.is_back_edge (src, pt) :
        assert bd_src is not None
        # position of the bound counter in the argument list of predicates
        bd_arg_pos = len (wto_comps) - 1
        # store inside PT for future use
        bd_pt.bd_arg_pos = bd_arg_pos

        pre_const = bd_src.pre_const (bd_arg_pos)
        post_const = bd_pt.post_const (bd_arg_pos)
        # re-entry (new iteration) into a wto component
        reentry_cons = list ()

        if spacer_args.args.up_counter:
          reentry_cons.append (0 <= pre_const)
          reentry_cons.append (post_const <= bd_pt.bval_const)
          reentry_cons.append (pre_const + 1 == post_const)
        else: # down counter
          reentry_cons.append (0 <= post_const)
          reentry_cons.append (post_const == pre_const - 1)
          reentry_cons.append (pre_const <= bd_pt.bval_const)

        #print 'reentry of', bd_pt.get_head (), reentry_cons
        bd_rule.add_bound_cons (reentry_cons)

    ### DONE WITH ALL INCOMING RULES

    # for every rule using pt, add exit constraints if needed
    for user,rule in hcp_utils.userRulesWalk (pt):
      bd_user = bd_pts [user]
      
      # get rule and create if needed
      bd_rule = bd_user.get_rule_from (bd_pt)
      if bd_rule is None :
        bd_rule = hcp.BoundedRule (rule.get_origin (), bd_user.head, bd_pt.head)
        bd_prob.add_rule (bd_rule)

      # check if user is in the innermost component of pt
      is_exiting = (len (wto_comps) > 0) and \
                     not hcp_utils.isInWtoComponent (user, wto_comps [-1])
      exit_cons = list ()
      if is_exiting :
        for i,comp in py_utils.reverse_enumerate (wto_comps):
          # shallow check - doesn't look at inner components
          if user not in comp :
            pre_const = bd_pt.pre_const (i)
            if not spacer_args.args.up_counter:
              exit_cons.append (pre_const == 0)
          else : break
      bd_rule.add_bound_cons (exit_cons)

  # set root, query for bd_prob
  bd_prob.set_root (bd_pts [orig_prob.root])
  bd_prob.set_query (bd_pts [orig_prob.query])

  # close the problem
  bd_prob.close ()

  return bd_prob
