import z3
import z3_utils as z3u
import hcp_utils
import hc_problem as hcp
import dfs
import py_utils
import stats
import spacer_args

from stats import time_stats

class FixedpointProblem (z3u.HelperBase):
    def __init__ (self, prob, bvals={}):
        z3u.HelperBase.__init__ (self, prob.ctx)
        self.prob = prob
        self.fp = None
        self.result = z3.unknown

        self.lemmas = dict ()
        self.cex = None

        # values of bound consts for pts
        #   partial map
        self.bvals = bvals

        # extracted hints for each pt
        self.hints = dict ()
        
        
    # TODO : look at other fp.set() options
    #        - print_certificate
    #        - generate_explanations
    #        - validate_result # now implemented by Nikolaj?
    #        - flexible_trace # to get the long cexes made possible by pdr??
    def solve (self, pp=False, z3validate=False):
        self.fp = z3.Fixedpoint (ctx=self.ctx)
        fp = self.fp
        fp.set (engine='pdr',
                use_farkas=True,
                generate_proof_trace=False)
        if not pp :
            print 'DISABLED PRE-PROCESSING'
            fp.set (slice=False,
                    inline_linear=False,
                    inline_eager=False)
        if z3validate: fp.set (validate_result=True)
        
        for pt in self.prob.pts: 
            fp.register_relation (pt.get_head ())

        for pt in self.prob.pts:
            pt.populate_fp (fp)
        
        if spacer_args.verbose >= 20:
            print 'Fixed Point Object is', fp.sexpr ()
        
        with stats.timer ('fp.query'), stats.block ("QUERY"):
            self.result = fp.query (self.prob.get_query_pred ())
    
        stats.put ('last_ind_lvl', fp.get_inductive_level ())
    
        #print 'Final FP is', fp

        if (self.result == z3.unsat):
            for pt in self.prob.pts:
                cover = fp.get_cover_delta (-1, pt.head)
                if z3.is_true (cover): self.lemmas [pt] = list ()
                else: self.lemmas [pt] = z3u.get_conjuncts (cover)
        elif (self.result == z3.sat):
            self.cex = fp.get_answer ()
        return self.result


    @time_stats
    def solve_bounded (self, pp=False, z3validate=False, hints={}, use_invs=False,
                       verbose=0):

        # whether to strengthen transition relation with invariants
        strengthen_trans = use_invs
        new_lemmas_only = use_invs

        ## old_fp != None if we are re-solving the same problem after making it stronger
        ## can re-use complete previous trace
        old_fp = self.fp
        
        # no hints and no invariants if re-solving.
        # we'll get everything from the previous state
        if old_fp is not None:
            use_invs = False
            hints = {}

        self.fp = z3.Fixedpoint (ctx=self.ctx)
        fp = self.fp
        fp.set (engine='pdr',
                use_farkas=True,
                generate_proof_trace=False)
        if not pp :
            print 'DISABLED PRE-PROCESSING'
            fp.set (slice=False,
                    inline_linear=False,
                    inline_eager=False)
        if z3validate: fp.set (validate_result=True)
        
        for pt in self.prob.pts:
            fp.register_relation (pt.get_head ())

        for pt in self.prob.pts:
            pt.populate_fp (fp, self.bvals.get (pt), use_invs, strengthen_trans)
            hints_pt = hints.get (pt)
            if hints_pt is not None:
              for h in hints_pt: fp.add_cover (0, pt.get_head (), h)
              
        if old_fp is not None:
            if verbose>1: print 'BEGIN: MOVE TRACE'
            for pt in self.prob.pts:
                decl = pt.get_head ()
                for lvl in range (old_fp.get_num_levels (decl)):
                    for c in z3u.conjuncts (old_fp.get_cover_delta (lvl, decl)):
                        if not z3.is_true (c):
                            if verbose>1: print 'Pred {} lvl {} lemma {}'.format (decl, lvl, c)
                            fp.add_cover (lvl, decl, c)
                for c in z3u.conjuncts (old_fp.get_cover_delta (-1, decl)):
                    if not z3.is_true (c):
                        if verbose>1: print 'Pred {} lvl oo lemma {}'.format (decl, c)
                        fp.add_cover (-1, decl, c)
            del old_fp
            old_fp = None
            if verbose>1: print 'END: MOVE TRACE'

        if verbose>10: print 'Fixed Point Object is', fp
        
        with stats.timer ('fp.query'), stats.lap('QUERY'):
          self.result = fp.query (self.prob.get_query_pred ())

        stats.put ('last_ind_lvl', fp.get_inductive_level ())

        if (self.result == z3.unsat):
            for pt in self.prob.pts:
                cover = fp.get_cover_delta (-1, pt.head)
                if z3.is_true (cover):
                  self.lemmas [pt] = list ()
                elif new_lemmas_only:
                  self.lemmas [pt] = list ()
                  for conj in z3u.get_conjuncts (cover):
                    if not pt.is_inv (conj): self.lemmas [pt].append (conj)
                else: self.lemmas [pt] = z3u.get_conjuncts (cover)
        elif (self.result == z3.sat):
            self.cex = fp.get_answer ()


        if verbose>0 and (self.result == z3.unsat):
          print 'BEGIN: HINTS'
          for pt in self.prob.pts:
            print pt,':'
            pt_hints = hints.get (pt)
            if pt_hints is not None:
              for h in pt_hints: print h
          print 'END: HINTS'

          print 'BEGIN: HINTS turned INVARS'
          for pt in self.prob.pts:
            print pt,':',
            pt_lemmas = self.lemmas [pt]
            pt_hints = hints.get (pt)
            cnt = 0
            if pt_hints is None: total = 0
            else:
              total = len (pt_hints)
              for h in pt_hints:
                for l in pt_lemmas:
                  if l.eq (h) : cnt += 1; break
            print '{} / {}'.format (cnt, total)
          print 'END: HINTS turned INVARS'

          print 'FP NUM LEVELS:', self.fp.get_num_levels (self.prob.get_query ().head)

        return self.result

    def print_traces (self) :
      fp = self.fp
      for pt in self.prob.pts:
        print 'Trace for', pt
        n = fp.get_num_levels (pt.head)
        for i in range (n):
          print '{} : {}'.format (i, fp.get_cover_delta (i, pt.head))
        print 'oo : {}'.format (fp.get_cover_delta (-1, pt.head))

    def get_lemmas (self, pt): return self.lemmas.get (pt)

    def get_cex (self) : return self.cex
    
    @time_stats
    def validate (self):
        if self.result == z3.unsat:
            return self.validate_unsat ()
        elif self.result == z3.sat:
            return self.validate_sat ()
        return None

    @time_stats
    def validate_sat (self):
        cex = list (z3u.conjuncts (self.cex))
        ### cex only contains the query
        if len (cex) <= 1: return True
        
        ### remove query
        cex = cex[1:]
        ### append placeholder for root
        cex.append (None)
        
        import itertools
        a, b = itertools.tee(reversed (cex))
        next (b, None)

        # get prob solver
        s, __tmp__ = self.prob.get_solver (get_assumps=False)

        pre_model = []
        for pre, post in itertools.izip (a, b):
            if pre is None:
                pre_pt = self.prob.get_root ()
                pre_decl = pre_pt.get_head ()
            else:
                pre_lemmas = z3u.conjuncts (pre)
                pre_decl = next(pre_lemmas, None).decl ()
                pre_pt = self.prob.get_pt (pre_decl)
                del pre_lemmas

            assert pre_pt.get_head ().eq (pre_decl)
            
            assert post is not None
            post_lemmas = z3u.conjuncts (post)
            post_decl = next(post_lemmas, None).decl ()
            post_pt = self.prob.get_pt (post_decl)
            
            assert post_pt.get_head ().eq (post_decl)

            # pick the rule
            r = post_pt.get_rule_from (pre_pt)
            assert r is not None
            r_tag = r.get_tag ()
            
            with z3u.pushed_solver (s):
                # pick post_pt
                s.add (post_pt.get_tag ())
                # pick rule
                s.add (r_tag)

                # pre-condition
                for lem in pre_model: s.add (lem)

                # post-condition
                and_post_lemmas = self.mk_and (*post_lemmas)
                s.add (post_pt.mk_post_lemma (and_post_lemmas))
                                                   
                #print 'Edge: {} -> {}'.format (pre, post)
                # print 'SOLVER BEGIN'
                # print s
                # print 'SOLVER END'

                # validate abstract result for pre --r--> post
                res = s.check (*r.get_used_assumps ())
                if res != z3.sat: return False
             
                ## assume res == z3.sat
                print 'MODEL for', post_decl
                s_model = s.model ()
                # reset pre_model
                pre_model[:] = []
                for i in range (post_decl.arity ()):
                    arg = post_pt.post_const (i)
                    val = s_model.eval (arg, model_completion=True)

                    var = post_pt.pre_const (i)
                    pre_model.append (var == val)
                #print pre_model
                print 'MODEL END'
                    
        return True

    def mk_bval_cons (self, pt):
      if pt.bval_const is None: return None
      if pt not in self.bvals:
        assert len (self.bvals) == 0
        return None
      return (pt.bval_const == z3.RealVal (self.bvals [pt], ctx=self.ctx))

    @time_stats
    def validate_unsat (self):
        """validate against the concrete transition relation"""
        s, __tmp__ = self.prob.get_solver (get_assumps = False)
        trans_assumps = self.prob.get_all_assumps ()
        
        with z3u.pushed_solver (s):
            # add all tranistion assumptions
            s.add (*trans_assumps)

            ### bound constraints
            for pt in self.prob.pts:
                bval_cons = self.mk_bval_cons (pt)
                if bval_cons is not None : s.add (bval_cons)
    
            ### pre-condition
            for pt in self.prob.pts:
                for l in self.get_lemmas (pt):
                    pre_lemma = pt.mk_pre_lemma (l)
                    was_at_pt = pt.get_pre_cond ()
                    s.add (z3.Implies (was_at_pt, pre_lemma))
            
            ### post-condition
            post = []
    
            for pt in self.prob.pts:
                for l in self.get_lemmas (pt):
                    post_lemma = pt.mk_post_lemma (l)
                    post.append (z3.And (pt.get_tag (), z3.Not (post_lemma)))
            ### safety
            q = self.prob.query
            post.append (q.get_tag ())
            
            post_exp = self.mk_or (*post)
            s.add (post_exp)

            res = s.check ()

        return res == z3.unsat

    def mk_pre_assumed_lemmas (self, pt, maker):
      was_at_pt = pt.get_pre_cond ()
      assumed_lemmas = []
      assumps = z3.AstVector (ctx=self.ctx)
      for l in self.get_lemmas (pt):
        a = maker (pt, l, variant=0)
        pre_lemma = pt.mk_pre_lemma (l)
        al = z3.Implies (a, pre_lemma)
        assumed_lemmas.append (z3.Implies (was_at_pt, al))
        assumps.push (a)
      return assumed_lemmas, assumps

    def mk_post_assumed_lemmas (self, pt, maker):
      assumed_lemmas = []
      assumps = z3.AstVector (ctx=self.ctx)
      for l in self.get_lemmas (pt):
        a = maker (pt, l, variant=1)
        post_lemma = pt.mk_post_lemma (l)
        al = z3.And (z3.Not (a), z3.Not (post_lemma))
        assumed_lemmas.append (z3.And (pt.get_tag (), al))
        assumps.push (a)
      return assumed_lemmas, assumps

    def _mk_hints (self, pt, lemma):
      """extract some expression in terms of the signature of pt.origin

      for now, substitute extra parameters in new signature with 0
      """
      res = list()

      new_lemma = pt.elim_bounds (lemma)

      if not z3.is_false (new_lemma):
          res.append (new_lemma)
       

      if spacer_args.args.up_counter and not new_lemma.eq (lemma):
          res.append (lemma)

      
      if len(res) > 0:
          print 'HINT:', pt.get_head (), res
      return res
    

    def add_hint (self, pt, *hints):
      l = self.hints.get (pt)
      if l is None:
        l = list ()
        self.hints [pt] = l
      l.extend (hints)

    def get_hints_pt (self, pt): return self.hints.get (pt)
    def get_hints (self): return self.hints

    @time_stats
    def refine (self):
        ''' Does refinement step. Returns True if refinement was
        successful (i.e., cex is spurious), and False otherwise. ''' 

        #print 'Cex is', self.cex
        if z3.is_and (self.cex):
            print 'CEX SZ:', self.cex.num_args ()
        
        cex = list (z3u.conjuncts (self.cex))
        ### cex only contains the query
        if len (cex) <= 1: return False
        
        ### remove query
        cex = cex[1:]
        ### append placeholder for root
        cex.append (None)
        
        import itertools
        a, b = itertools.tee(reversed (cex))
        next (b, None)

        s = z3.Solver (ctx=self.ctx)

        rules = set ()
        assumps = list ()

        for step, (pre,post) in enumerate (itertools.izip (a, b), start=1):
            ### get pts, decls for pre,post
            ###   and lemmas for post

            if pre is None:
                pre_pt = self.prob.get_root ()
                pre_decl = pre_pt.get_head ()
            else:
                pre_lemmas = z3u.conjuncts (pre)
                pre_decl = next(pre_lemmas, None).decl ()
                pre_pt = self.prob.get_pt (pre_decl)
                del pre_lemmas

            assert pre_pt.get_head ().eq (pre_decl)
            
            assert post is not None
            post_lemmas = z3u.conjuncts (post)
            post_decl = next(post_lemmas, None).decl ()
            post_pt = self.prob.get_pt (post_decl)
            
            assert post_pt.get_head ().eq (post_decl)

            ### constraint for rule between pre and post
            r = post_pt.get_rule_from (pre_pt)
            assert r is not None
            s.add (*r.mk_assumed_trans_variant (step))


            ### update assumps
            if r not in rules:
              rules.add (r)
              assumps.extend (r.get_all_assumps ())

            
            ###
            ### 2 reasons to not enable post_lemmas
            ###   1. with python interface, this is really, really, slow
            ###   2. leads to more robust refinement
            ###
            # ### post-lemmas
            # for lem in post_lemmas:
            #   s.add (hcp.mk_post_lemma_variant (post_decl, lem, step))


        core = None
        # validate concrete result
        with stats.timer ('fp_problem.refine.check'):
            if spacer_args.args.refine_use_mus:
                res, core = hcp_utils.mus_ltf (s, *assumps)
            else:
                res = s.check (*assumps)
                if res == z3.unsat: core = s.unsat_core ()
            
        
        if core is not None:
            print '\t REFINEMENT WITH', len(core)
            for r in rules:
              r.refine_with_assumps (core)
            return True

        assert res == z3.sat
        return False
    
    @time_stats
    def refine_naive (self):
        ''' Does refinement step. Returns True if refinement was
        successful (i.e., cex is spurious), and False otherwise. ''' 

        #print 'Cex is', self.cex
        if z3.is_and (self.cex):
            print 'CEX SZ:', self.cex.num_args ()
        
        cex = list (z3u.conjuncts (self.cex))
        ### cex only contains the query
        if len (cex) <= 1: return False
        
        ### remove query
        cex = cex[1:]
        ### append placeholder for root
        cex.append (None)
        
        import itertools
        a, b = itertools.tee(reversed (cex))
        next (b, None)

        pre_model = []
        for pre, post in itertools.izip (a, b):
            if pre is None:
                pre_pt = self.prob.get_root ()
                pre_decl = pre_pt.get_head ()
            else:
                pre_lemmas = z3u.conjuncts (pre)
                pre_decl = next(pre_lemmas, None).decl ()
                pre_pt = self.prob.get_pt (pre_decl)
                del pre_lemmas

            assert pre_pt.get_head ().eq (pre_decl)
            
            assert post is not None
            post_lemmas = z3u.conjuncts (post)
            post_decl = next(post_lemmas, None).decl ()
            post_pt = self.prob.get_pt (post_decl)
            
            assert post_pt.get_head ().eq (post_decl)

            # get post solver
            s, __tmp__ = self.prob.get_solver (get_assumps=False)
            
            # pick the rule
            r = post_pt.get_rule_from (pre_pt)
            assert r is not None
            r_tag = r.get_tag ()
            
            with z3u.pushed_solver (s):
                # pick post_pt
                s.add (post_pt.get_tag ())
                # pick rule
                s.add (r_tag)

                # pre-condition
                for lem in pre_model: s.add (lem)
                
                # post-condition
                and_post_lemmas = self.mk_and (*post_lemmas)
                s.add (post_pt.mk_post_lemma (and_post_lemmas))

                #print 'Edge: {} -> {}'.format (pre, post)
                # print 'SOLVER BEGIN'
                # print s
                # print 'SOLVER END'

                # validate abstract result for pre --r--> post
                # print 'Validation with:', r.get_used_assumps ()
                # res = s.check (*r.get_used_assumps ())
                # assert res == z3.sat


                # validate concrete result
                res = s.check (*r.get_all_assumps ())

                if res != z3.sat:
                    core = s.unsat_core ()
                    print '\t REFINEMENT WITH', core
                    r.refine_with_assumps (core)
                    return True
                else:
                    print 'MODEL for', post_decl
                    s_model = s.model ()
                    # reset pre_model
                    pre_model[:] = []
                    for i in range (post_decl.arity ()):
                        arg = post_pt.post_const (i)
                        val = s_model.eval (arg, model_completion=True)

                        var = post_pt.pre_const (i)
                        pre_model.append (var == val)
                    #print pre_model
                    print 'MODEL END'
                    
        return False


    @time_stats
    def bounded_abstract (self) :

      # maps from pt to AssumedLemmas objects
      pre_assumps = dict ()
      post_assumps = dict ()

      def mk_assumption (pt, f, variant=0):
        """makes an assump. var. and stores in the maps pre/post-assumps
        """
        if variant == 0: d = pre_assumps
        else: d = post_assumps
        al = d.get (pt)
        if al is None :
          al = hcp_utils.AssumedLemmas (self.prob.mk_assume_var, ctx=self.ctx)
          d [pt] = al
        return al.assume (f)

      if spacer_args.args.conc_abs:
          s, __tmp__ = self.prob.get_solver (get_assumps=False)
          trans_assumptions = self.prob.get_all_assumps ()
      else:
          s, trans_assumptions = self.prob.get_solver ()

      with z3u.pushed_solver (s):
          post_a = z3.AstVector (ctx=self.ctx)
          pre_a = z3.AstVector (ctx=self.ctx)

          ### pre and post lemmas
          post_lemmas = []
          for pt in self.prob.pts:
            al, assumps = self.mk_pre_assumed_lemmas (pt, mk_assumption)
            s.add (*al)
            z3u.extend (pre_a, assumps)

            al, assumps = self.mk_post_assumed_lemmas (pt, mk_assumption)
            post_lemmas.extend (al)
            z3u.extend (post_a, assumps)

          q = self.prob.get_query ()
          post_lemmas.append (q.get_tag ())
          s.add (self.mk_or (*post_lemmas))

          all_assumptions = list(trans_assumptions)


          if spacer_args.args.fix_lemmas:
              s.add (*map (z3.Not, post_a))
              s.add (*pre_a)
          else:
              all_assumptions.extend (map(z3.Not, post_a))
              all_assumptions.extend (pre_a)
                 

          for pt in self.prob.pts:
              bval_cons = self.mk_bval_cons (pt)
              if bval_cons is not None : s.add (bval_cons)

          res, used_assump = hcp_utils.mus_ltf (s, *all_assumptions)
          assert res == z3.unsat
          #res = s.check (*all_assumptions)
          #assert res == z3.unsat
          #used_assump = s.unsat_core ()
      
      self.prob.abstract_by_assumptions (used_assump)


    @time_stats
    def extract_unbounded_invs (self, use_invs=True) :
      # assume lemmas are already validated
      #  (may add a class field in future for this purpose)

      # maps from pt to AssumedLemmas objects
      pre_assumps = dict ()
      post_assumps = dict ()

      def mk_assumption (pt, f, variant=0):
        """makes an assump. var. and stores in the maps pre/post-assumps
        """
        if variant == 0: d = pre_assumps
        else: d = post_assumps
        al = d.get (pt)
        if al is None :
          al = hcp_utils.AssumedLemmas (self.prob.mk_assume_var, ctx=self.ctx)
          d [pt] = al
        return al.assume (f)

      def post_to_pre (assumps) :
        """return pre assumption var.s for assumps, using post/pre-assumps maps
        """
        to_ret = z3.AstVector (ctx=self.ctx)
        for a in assumps:
          for pt in self.prob.pts:
            al_post = post_assumps.get (pt)
            al_pre = pre_assumps.get (pt)
            if al_post is None: continue
            l = al_post.get_lemma (a)
            if l is None: continue
            to_ret.push (al_pre.get_assump (l))
            break
        return to_ret

      s, __tmp__ = self.prob.get_solver (get_assumps=False)
      trans_assumps = self.prob.get_all_assumps ()

      with z3u.pushed_solver (s):
        # add all transition assumptions
        s.add (*trans_assumps)

        post_a = z3.AstVector (ctx=self.ctx)
        pre_a = z3.AstVector (ctx=self.ctx)
  
        ### pre and post lemmas
        post_lemmas = []
        for pt in self.prob.pts:
          al, assumps = self.mk_pre_assumed_lemmas (pt, mk_assumption)
          s.add (*al)
          z3u.extend (pre_a, assumps)
  
          al, assumps = self.mk_post_assumed_lemmas (pt, mk_assumption)
          post_lemmas.extend (al)
          z3u.extend (post_a, assumps)
  
        s.add (self.mk_or (*post_lemmas))
  
        # union of all mus'es
        muses = z3.AstVector (ctx=self.ctx)

        ### iterate until convergence
        count = 0
        while True :
          print 'Iteration', count
          count += 1
  
          true_lits = z3.AstVector (ctx=self.ctx)
          # true_lits = pre_a + muses
          z3u.extend (true_lits, pre_a)
          z3u.extend (true_lits, muses)
  
          # get new mus
          mus = hcp_utils.min_unsat_subset (s, true_lits, post_a)
          if len (mus) == 0: break
          # union with muses
          z3u.extend (muses, mus)
  
          # get complement of mus in post_a
          new_post_a = z3.AstVector (ctx=self.ctx)
          for a in post_a:
            if a not in mus: new_post_a.push (a)
          post_a = new_post_a
  
          # new pre_a
          pre_a = post_to_pre (post_a)

      # store unbounded invariants
      for pt in self.prob.pts :
        al = pre_assumps.get (pt)
        if al is None: continue
        for l in self.get_lemmas (pt):
          a = al.get_assump (l)
          if a in pre_a: print l; pt.add_inv (l, add_to_solver=use_invs)
          else: self.add_hint (pt, *self._mk_hints (pt, l))

    def converged (self):
      """true iff False is an invariant of query"""
      q = self.prob.query
      for inv in q.get_invs ():
        if z3.is_false (inv): return True
      return False
