#! /usr/bin/env python

import sys
import spacer_args

import z3
import z3_utils
import dfs
import hc_problem as hcp
import fp_problem as fpb
import bounded_exploration as bd_exp
import hcp_utils
import stats




def solve (prob,
           pp=False, z3validate=False, verbose=0,
           init_bound=0, bd_increment=1,
           validate=True, use_invs=True, use_hints=True,
           pba=False, cba=False, refine_naive=False,
           trace=False, proof_out=False):

  stats.put ('Result', 'UNKNOWN')

  bvals = dict ()
  bound = init_bound
  for pt in prob.pts : bvals [pt] = bound

  num_iters = 0
  refine_cnt = 0

  fp_prob = fpb.FixedpointProblem (prob, bvals)
  hints = {}
  while True :
    stats.put ('Iterations', num_iters)
    stats.put ('bound', bound)
    stats.brunch_print ()


    fp_res = fp_prob.solve_bounded (pp=pp, z3validate=z3validate, 
                                    hints=hints, use_invs=use_invs,
                                    verbose=spacer_args.verbose)
    num_iters += 1
    print 'Result is: ', fp_res
    if pba:
      print 'Current Abstraction'
      for pt in prob.pts:
        ua = pt.get_used_assumps ()
        aa = pt.get_all_assumps ()
        print '{0:30}{1}\t{2:>3}/{3:}'.format (pt, bvals[pt], len(ua), len(aa))
        #print ua
      print 'End'

    if fp_res == z3.sat :
      if spacer_args.verbose>10: print 'Counterexample :', fp_prob.get_cex ()
      if validate:
        if not fp_prob.validate ():
          stats.put('ValFail', 'CEX')
          break
      if (not pba) or prob.is_concrete ():
        ### cex is real
        stats.put ('Result', 'CEX')
        break
      if cba and (spacer_args.args.max_refine == 0 or \
                  refine_cnt < spacer_args.args.max_refine):
        # refine the abstraction
        with stats.block('REFINEMENT'):
          if refine_naive: refined = fp_prob.refine_naive ()
          else: refined = fp_prob.refine ()
          if not refined:
            stats.put ('Result', 'CEX')
            break
          refine_cnt += 1
      else:
        # concretize the abstraction and re-solve
        print 'Concretizing...'
        prob.concretize ()
    elif fp_res == z3.unsat :
      # reset refinement counter
      refine_cnt = 0

      if spacer_args.verbose>0:
        for pt in prob.pts: 
          print 'Lemmas at {} are {}'.format (pt, fp_prob.get_lemmas (pt))
      if trace: fp_prob.print_traces ()
      if validate:
        if not fp_prob.validate ():
          stats.put ('ValFail', 'SAFE')
          break
      
      fp_prob.extract_unbounded_invs (use_invs=use_invs)

      if spacer_args.verbose>-1:
        print 'Unbounded Invariants:'
        for pt in prob.pts:
          print pt,':'
          print pt.invs

      if fp_prob.converged ():
        stats.put ('Result', 'SAFE')
        break

      if pba:
        with stats.block ('ABSTRACTION'):
          fp_prob.bounded_abstract ()

      # prepare for next fixedpoint problem

      if use_hints: hints = fp_prob.get_hints ()
      bound += bd_increment
      # increment bound vals
      for pt in prob.pts: bvals [pt] = bound
      # next fixedpoint problem
      fp_prob = fpb.FixedpointProblem (prob, bvals)

    else:
      assert fp_res == z3.unknown
      print 'Unable to solve the bounded problem'
      break

  stats.put ('Iterations', num_iters)

  if proof_out != None:
    f = open (proof_out, 'w')
    f.write (fp.get_answer().sexpr())
    f.close()


def solve_vanilla (prob,
                   pp=False,
                   z3validate=False, validate=False,
                   verbose=0, trace=False):
  print 'VANILLA'
  stats.put ('Result', 'UNKNOWN')

  fp_prob = fpb.FixedpointProblem (prob)
  fp_res = fp_prob.solve (pp=pp, z3validate=z3validate)
  
  if fp_res == z3.sat :
    if spacer_args.verbose>10:
      c = fp_prob.get_cex ()
      print 'Counterexample :', c
    stats.put ('Result', 'CEX')
  elif fp_res == z3.unsat :
    if spacer_args.verbose>10:
      for pt in prob.pts: 
        print 'Lemmas at {} are {}'.format (pt, fp_prob.get_lemmas (pt))
      fp_prob.print_traces ()
    stats.put ('Result', 'SAFE')
  if validate:
    print 'Validating...'
    print fp_prob.validate ()
  if trace: fp_prob.print_traces ()

@stats.time_stats
def initialize (filename, pp, dummy_query) :
  """create horn clause problem from input file and return a bounded version by
bounding the predicates and rules

  returned problem is just a skeleton with variables standing for concrete bound
values
  """
  parsing_ctx = z3.Context ()
  bounding_ctx = parsing_ctx

  prob = hcp.create_hc_problem (filename, pp=pp,
                                z3ctx=parsing_ctx, dummy_query=dummy_query)
  prob.close ()
  if spacer_args.verbose > 10 : print prob


  ## bound the problem
  bd_prob = bd_exp.create_bounded_hc_problem (prob, z3.RealSort (ctx=bounding_ctx))
  assert bd_prob.closed

  if spacer_args.verbose > 10: print bd_prob

  return (prob, bd_prob)


@stats.time_stats
def main (argv):
  import z3types
  args = spacer_args.parseArgs (argv[1:])

  # set z3 global options
  z3.set_option (verbose=args.z3verbosity)

  if args.z3_trace is not None:
    for t in args.z3_trace.split (':'):
      z3.enable_trace (t)

  #z3.enable_trace ('pdr')
  #z3.enable_trace ('dl')
  #z3.enable_trace ('smt_relation')

  (prob, bd_prob) = initialize (args.filename, False, args.dummy_query)

  print 'Initialization done'

  try:
    with stats.timer ('MAIN.solve'):
      if args.vanilla: solve_vanilla (bd_prob,
                                      pp=args.pp,
                                      z3validate=args.z3validate, validate=args.validate,
                                      verbose=args.verbose, trace=args.trace)
      else:
        solve (bd_prob,
               pp=args.pp, z3validate=args.z3validate, verbose=args.verbose,
               init_bound=args.init_bound, bd_increment=args.bd_increment,
               validate=args.validate, use_invs=args.use_invs, use_hints=args.use_hints,
               pba=args.pba, cba=args.cba, refine_naive=args.refine_naive,
               trace=args.trace, proof_out=args.proof_out)
  except:
    print 'EXCEPTION'
    import traceback
    traceback.print_exc(file=sys.stdout)
    return 3
  finally:
    stats.brunch_print ()

  return 0


if __name__ == '__main__':
  # unbuffered output
  import os
  sys.stdout = os.fdopen (sys.stdout.fileno (), 'w', 0)  
  sys.exit (main(sys.argv))
